
(* what happens to a particular vote *)

structure Find = struct 
local open Types ;
in 

datatype votedest = DExcl of {excl : int, now_for : int} 
		    (* candidate excluded, vote transferred to another *)
		  | DElec of int (* candidate elected *)
		  | DElecThis of int  (* candidate elected using this vote *)
		  | DElecTfr of {elec : int, now_for : int, tv : tv} 
		    (* vote transferred upon election of candidate *)
		  | DInit of int ; (* initial count *)

(* when vote exhuasted, now_for given as 0 *)

(* countedFor : cand -> votedest -> bool
  test whether destination indicates that vote was counted for given candidate
  (or, for cand=0, was exhausted) *)
fun countedFor cand (DExcl {now_for, ...}) = now_for = cand  
  | countedFor cand (DElecTfr {now_for, ...}) = now_for = cand   
  | countedFor cand (DInit now_for) = now_for = cand   
  | countedFor cand _ = false ;

(* given list of non-continuing candidates, delete preferences for those
  and find next continuing candidate in vote,
  returning rest of vote after that candidate *)
  
(* destAtCount : int -> (count * elecEvent) list -> 
  count -> string -> cand option
  nc = number of candidates
  events = list of events, in increasing order of count
    (ie original order, reversed) - can filter out those not used
  str = vote as string of bytes
  count = number of count of interest
  returns SOME cand if vote counted at count for cand, otherwise NONE *)

fun destAtCount nc ((1, InitDecd decd) :: events) count str = 
  let val sstr = Substring.all str ;
    (* ctga indicates whether candidate continuing *)
    val ctga = Array.array (nc + 1, true) ;
    (* set ctga to false for deceased candidates *)
    fun notctg dc = Array.update (ctga, dc, false) ;
    val () = app notctg decd ;
    fun isctg cand = Array.sub (ctga, cand) ;
    val (c, subv) = Count.getpref isctg sstr ; (* first count *)
    fun ne (r1, r2) = (tvcompare (r1, r2) <> EQUAL) ; infix ne ;

    (* destr : (''count * elecEvent) list -> cand * ''count * tv * substring -> 
      ''count -> cand option
      if arg 2 is (curr, lcnt, subv),
      gets destination of remainder of vote, where vote is currently for curr,
      last counted at count lcnt, and subv is rest of vote (after curr) *)
    fun destr _ (cand, _, _, _) 1 = SOME cand
      | destr _ (0, _, _, _) _ = NONE (* excluded at earlier count *)
      | destr (evs as (cnt, _) :: _) cls count = 
	if count < cnt then NONE else destr' evs cls count
      | destr [] _ _ = NONE

    and destr' ((cnt, ExclDet cand) :: events) cls count = (
        Array.update (ctga, cand, false) ;
        destr' events cls count)

      | destr' ((cnt, ExclDist (cand, tvc, _)) :: events) 
	  (cls as (curr, lcnt, tvv, subv)) count = (
	if cand <> curr orelse tvc ne tvv then destr events cls count
	else let val (curr', subv') = Count.getpref isctg subv 
	  in if cnt = count then SOME curr'
	    else destr events (curr', cnt, tvv, subv') count end)

      | destr' ((cnt, Elect (cand :: es)) :: events) 
	  (cls as (curr, lcnt, _, _)) count = (
        Array.update (ctga, cand, false) ;
	if cand <> curr then destr' ((cnt, Elect es) :: events) cls count
	(* see if vote counted at this count and so will be recounted
	  when candidate's surplus distributed *)
	else if lcnt = cnt then (* vote counted at this count *)
	  destr' ((cnt, Elect es) :: events) cls count
	else NONE) (* vote counted for candidate at a previous count *)

      | destr' ((cnt, Elect []) :: events) cls count = destr events cls count

      | destr' ((cnt, Surplus (cand, tv, scnt)) :: events) 
	  (cls as (curr, lcnt, tvv, subv)) count = 
	if cand <> curr then destr events cls count
	else let val (curr', subv') = Count.getpref isctg subv 
	  in if cnt = count then SOME curr'
	    else destr events (curr', cnt, tv, subv') count end 
	    
      | destr' (event :: events) csubv count = destr events csubv count 

  in destr events (c, 1, TV (1, 1), subv) count end ;

(* dest : int -> (count * elecEvent) list -> string -> (count * votedest) list
  nc = number of candidates
  events = list of events, in increasing order of count
    (ie original order, reversed) - can filter out those not used
  str = vote as string of bytes
  returns list of destinations of vote *)

fun dest nc ((1, InitDecd decd) :: events) str = 
  let val sstr = Substring.all str ;
    (* ctga indicates whether candidate continuing *)
    val ctga = Array.array (nc + 1, true) ;
    (* set ctga to false for deceased candidates *)
    fun notctg dc = Array.update (ctga, dc, false) ;
    val () = app notctg decd ;
    fun isctg cand = Array.sub (ctga, cand) ;
    val (c, subv) = Count.getpref isctg sstr ;
    fun ne (r1, r2) = (tvcompare (r1, r2) <> EQUAL) ; infix ne ;

    (* destr : (''count * elecEvent) list -> cand * ''count * tv * substring -> 
      (''count * votedest) list
      if arg 2 is (curr, lcnt, subv),
      gets destination of remainder of vote, where vote is currently for curr,
      last counted at count lcnt, with transfer value tv,
      and subv is rest of vote (after curr) *)

    fun destr _ (0, _, _, _) = [] (* vote already exhausted *)

      | destr ((cnt, ExclDet cand) :: events) cls = (
        Array.update (ctga, cand, false) ;
        destr events cls)

      | destr ((cnt, ExclDist (cand, tvc, _)) :: events) 
	  (cls as (curr, lcnt, tvv, subv)) = (
	if cand <> curr orelse tvc ne tvv then destr events cls  
	else let val (curr', subv') = Count.getpref isctg subv 
	  in (cnt, DExcl {excl = cand, now_for = curr'}) ::
	      destr events (curr', cnt, tvv, subv') end)

      | destr ((cnt, Elect (cand :: es)) :: events) 
	  (cls as (curr, lcnt, _, _)) = (
        Array.update (ctga, cand, false) ;
	if cand <> curr then destr ((cnt, Elect es) :: events) cls  
	(* see if vote counted at this count and so will be recounted
	  when candidate's surplus distributed *)
	else if lcnt = cnt then (* vote counted at this count *)
	  (cnt, DElecThis cand) :: destr ((cnt, Elect es) :: events) cls  
	(* vote counted for candidate at a previous count,
	  so we are finished with such a vote (++) *)
	else [(cnt, DElec cand)]) 

      | destr ((cnt, Elect []) :: events) cls = destr events cls

      | destr ((cnt, Surplus (cand, tv, scnt)) :: events) 
	  (cls as (curr, lcnt, tvv, subv)) = 
	(* note - votes for cand which are not to be transferred 
	  are already out of the picture, see (++) above *)
	if cand <> curr then destr events cls
	else 
	 let val (curr', subv') = Count.getpref isctg subv 
	  in (cnt, DElecTfr {elec = curr, now_for = curr', tv = tv}) ::
	    destr events (curr', cnt, tv, subv') end 
	    
      | destr (event :: events) csubv = destr events csubv 
      | destr [] _ = [] ;

  in (1, DInit c) :: destr events (c, 1, TV (1, 1), subv) end ;

(* coll : (count * votedest -> bool) -> int -> (count * elecEvent) list ->
  string list -> (count * votedest) list list
  for each vote, get destination list, and 
  just return destinations satisfying f *) 
fun coll f nc events votes = 
  (map (List.filter f o dest nc events) votes) ;
  
(* collvotes : (count * votedest -> bool) -> int -> (count * elecEvent) list ->
  ('voteId * string) list -> ('voteId * string) list
  filter vote list according to whether dest of vote contains entry sat'g f,
  note - here vote is pair of string and vote id *)
fun collvotes f nc events votes =
  let fun ctf (vid, vote) = List.exists f (dest nc events vote) ;
  in List.filter ctf votes end ;
  
(* findcc : int -> int -> int -> (count * elecEvent) list -> 
  ('voteId * string) list -> ('voteId * string) list
  find votes counted to candidate cand at count count,
  note - here vote is pair of string and vote id *)
fun findcc nc cand count events votes = 
  let fun f (vid, str) = case destAtCount nc events count str of NONE => false
	| SOME cand' => cand = cand' ;
  in List.filter f votes end ;

(* cand : ('count * votedest) -> int option
  find the new candidate in a (cnt, votedest) pair which
  gives the new candidate when the vote is counted *)
fun cand (_, DElecTfr {now_for, ...}) = SOME now_for
  | cand (_, DExcl {now_for, ...}) = SOME now_for
  | cand (_, DInit cand) = SOME cand
  | cand (_, _) = NONE ;
  
(* candnums : int -> ('count * votedest) list -> (cand * int) list
  given a list of (cnt, votedest) pairs which reflect the vote being counted,
  tally up the number of these votes for each candidate;
  results are (candidate, no of votes) pairs;
  we filter out zero results *)
fun candnums nc dests = 
  let val cntd = Array.array (nc + 1, 0) ;

    (* count_by_cand : ('count * votedest) list -> unit *)
    fun count_by_cand [] = ()
      | count_by_cand (cd :: cds) = (
	case cand cd of SOME sc => updatef cntd sc (fn i => i + 1)
	  | NONE => () ;
	count_by_cand cds) ;
    
  in count_by_cand dests ; 
    List.filter (fn (_, n) => n <> 0) (numl 0 (toList cntd)) end ; 

(* candnums' :
  int -> (count * elecEvent) list -> int -> string list -> (cand * int) list
  get results of count count using destAtCount *)
fun candnums' nc events count votes = 
  let val cntd = Array.array (nc + 1, 0) ;

    fun count_by_cand [] = ()
      | count_by_cand (str :: strs) = (
	case destAtCount nc events count str of NONE => ()
	  | SOME cand => updatef cntd cand (fn i => i + 1) ;
	count_by_cand strs) ;

  in count_by_cand votes ; 
    List.filter (fn (_, n) => n <> 0) (numl 0 (toList cntd)) end ; 
    
(** following is a test which examines all votes using both the 
  routines dest and destAtCount to determine the numbers counted for
  each candidate at a given count;
  result is (count number, whether dest and destAtCount give same result,
    whether dest and results in info correspond);
  **)

infix eqex0 ;

fun del0 l = List.filter (fn (_, n) => n <> 0) l ;

fun l1 eqex0 l2 = del0 l1 = del0 l2 ;

(* test : info -> string list -> int -> int * bool * bool
  do test described above for a particular count *)
fun test info votes count = 
  let 
    val Info {events = ref evs, results = ref results,
      fixed = {nc, ...}, ...} = info ;
    val events = rev evs ;
    fun f (cnt, dest) = cnt = count ;
    val ctic = coll f nc events votes ; (* uses function dest *)
    val cticcs = (candnums nc (List.concat ctic)) ;
    val cticcs' = (candnums' nc events count votes) ; (* uses destAtCount *)
    val SOME (_, Cres {count = (_, cl), ...}) =
      List.find (fn (i, _) => i = count) results ;
  in (count, cticcs eqex0 cticcs', cticcs eqex0 cl) end ;
  
fun upto (m, n) = if m > n then [] else m :: upto (m+1, n) ;

(* testall : info -> string list -> (int * bool * bool) list
  do test described above for a all counts, return errors *)
fun testall info votes = 
  let 
    val Info {results = ref results, ...} = info ;
    val (lastcount, _) :: _ = results ;
    val ress = map (test info votes) (upto (1, lastcount)) ;
    val errors = List.filter (fn (_, true, true) => false | _ => true) ress ;
  in errors end ;

(** this routine actually returns the votes for the given candidate at the
  given count, and also finds them using both routines above, checking
  that the result is the same **)
fun findcc' (Info {events = ref evs, fixed = {nc, ...}, ...}) 
    cand count votes =
  let val events = rev evs ;
    val idvotes = Types.numl 1 votes ;
    fun v2il (id, v) = (id, RPV.s2ilvote v) ;
    (* findcc uses function destAtCount *)
    val vs = (findcc nc cand count events idvotes) ; 
    fun f (ct, dest) = ct = count andalso countedFor cand dest ;
    val vs' = (collvotes f nc events idvotes) ; (* uses function dest *)
    val check = vs = vs' ;
    val lens = map length [vs, vs'] ;
  in {cand = cand, count = count, lens = lens, 
      ivotes = map v2il vs, check = check} end ;

end ; (* local open Types ; *)
end ; (* structure Find *)

(* to find exhausted votes at count 1 due to deceased candidates *)
(*
fun f (1, DInit 0) = true | f _ = false ;
val ve1 = collvotes f nc events (Types.numl 1 votes) ;
fun g (cnt, vote) = destAtCount nc events 1 vote = SOME 0 ;
val ve2 = List.filter g (Types.numl 1 votes) ;
val true = ve1 = ve2 ;
*)

(* sample - find votes for given cand at given count 
val cand = 13 ; val count = 15 ;
val cand = 0 ; val count = 15 ;
val Types.Info {events = ref evs, ...} = info ;
val events = rev evs ;
val idvotes = Types.numl 1 votes ;
fun v2il (id, v) = (id, RPV.s2ilvote v) ;
val idvs = map v2il (Find.findcc nc cand count events idvotes) ;
fun f (ct, dest) = ct = count andalso countedFor cand dest ;
val idvs' = map v2il (collvotes f nc events idvotes) ;
val true = idvs = idvs' ;
val num = length idvs ;
*)

(* additionally, to look at history of these votes *)
(*
fun getDests (id, il) = 
  let val str = RPV.il2svote il ;
    val ds = dest nc events str ;
  in (id, il, ds) end ;

val idvds = map getDests idvs ;
*)

(* test for Brindabella 2001 
val count = 6 ; val cand = 9 ;
val Types.Info {events = ref evs, ...} = info ;
val events = rev evs ;
val idvotes = Types.numl 1 votes ;
fun v2il (id, v) = (id, RPV.s2ilvote v) ;
val idvs = map v2il (Find.findcc nc cand count events idvotes) ;
fun f (ct, dest) = ct = count andalso countedFor cand dest ;
val idvs' = map v2il (collvotes f nc events idvotes) ;
val true = idvs = idvs' ;
val num = length idvs ;
*)


