
(* process candidate tables provided by ACT Electoral Commission *)

structure Getcands = struct 

(* chkconsec : int -> int list -> bool
  check that numbers increase consecutively, starting from j = i + 1 *)
fun chkconsec i (j :: js) = j = i + 1 andalso chkconsec j js 
  | chkconsec i [] = true ;

(* procln : string -> 
  {ID : int, ccode : int, cname : string, ecode : int, pcode : int} *)
fun procln ln = 
  let val [fldCandIDstr,ecodestr,pcodestr,ccodestr,cname] =
      String.tokens (fn ch => ch = #",") ln ;
    val [fldCandID,ecode,pcode,ccode] = 
      map (Option.valOf o Int.fromString) 
        [fldCandIDstr,ecodestr,pcodestr,ccodestr] ;
    val crec = {ID = fldCandID, 
      ecode = ecode, pcode = pcode, ccode = ccode, cname = cname} ;
  in crec end ;

(* numcan : int -> 'a list list -> (int * 'a) list list
  assign candidate numbers for counting, starting from i *)
fun numcan i ((c :: cs) :: css) = 
    let val ncs :: ncss = numcan (i+1) (cs :: css) ;
    in ((i, c) :: ncs) :: ncss end 
  | numcan i ([] :: css) = [] :: numcan i css
  | numcan i [] = [] ;

(* format of tblCands.txt : 
fldCandID,ecode,pcode,ccode,cname
1,1,0,0,WILLIAMS Athol
2,1,0,1,WOOD Bill
*)
(* example
  val candfile = "../2001data/tblCands.txt" ; *)
fun getcands candfile = 
  let val () = print ("using candidate table " ^ candfile ^ "\n") ;
    local open TextIO in 
      val cands = openIn candfile ;
      val candstr = inputAll cands ;
      val _ = closeIn cands ;
    end ; (* local open TextIO *)
    val candlns = tl (String.tokens (fn ch => ch = #"\n") candstr) ;

    val plns = map procln candlns ;

    (* split into electorates, then parties *)
    val eplns = Gen.groupeq #ecode plns ;
    val ncands = map length eplns ; (* [27, 26, 41] *)
    val peplns = map (Gen.groupeq #pcode) eplns ;
    (* sort party groups by name, to correspond with published output *)
    val speplns = map (map (Listsort.sort 
	(Gen.trComp #cname String.compare))) peplns ;

    val nes = map (#ecode o hd) eplns ; 
    val true = (chkconsec 0) nes ;
    val npss = map (map (#pcode o hd)) peplns ;
    (* should each be [0, 1, 2, 3, 4, 5, 6, 7, ... ] *)
    val true = List.all (chkconsec ~1) npss ;
    val ncsss = map (map (map #ccode)) peplns ;
    val true = List.all (List.all (chkconsec ~1)) ncsss ;

    val nelns = map (numcan 1) speplns ;
    (* now revert to original ordering, ie sorted by ccode *)
    val cslns = map (map (Listsort.sort 
	(Gen.trComp (#ccode o #2) Int.compare))) nelns ;
    (* set vectors to look up candidate number by pcode, ccode *)
    (* for list of candidates within a party, set vector *)
    fun fst (a, b) = a ;
    fun canvec cs = Vector.fromList (map fst cs) ;
    fun gpvec gcs = Vector.fromList (map canvec gcs) ;
    (* one such vector for each electorate *)
    val evecs : int Vector.vector Vector.vector list = map gpvec cslns ;

    (* vectors for looking up candidate names *)
    val flatcls = map List.concat speplns ;
    val dummy = {ID = ~1, ccode = ~1,
      cname = "dummy record for candidate 0", ecode = ~1, pcode = ~1} ;
    val candvecs = map (fn rs => Vector.fromList (dummy :: rs)) flatcls ;  

    fun candinfo n = {evec = List.nth (evecs, n), ncand = List.nth (ncands, n),
      candvec = List.nth (candvecs, n)} ;
  in candinfo end ;
end ; (* structure Getcands *)

(* note, electorates are
Brindabella,1
Ginninderra,2
Molonglo,3
*)

