
(* process vote tables provided by ACT Electoral Commission *)

(*
fun ug _ = use "Gettbls.sml" ;

val _ = app load ["Int", "Listsort", "Gen", "Substring"] ;
*)

structure Getvotes = struct 

(* note, electorates are
Brindabella,1
Ginninderra,2
Molonglo,3
*)

(* format of votes file
fldID,batch,pindex,pref,ccode,pcode,rcand
1,0,11247,1,0,1,0
2,0,11247,2,1,1,0 
*)
(* procprf : substring -> int list
  process single line of votes (preferences) file *)
fun procprf sstr = 
  let
    (* ignore record ID and Robson rotation info *)
    val [_, batch_str, pindex_str, pref_str, ccode_str, pcode_str, _] = 
      Substring.tokens (fn ch => ch = #",") sstr ;
    val ifds as [batch, pindex, pref, ccode, pcode] = 
      map (Option.valOf o Int.fromString o Substring.string) 
	[batch_str, pindex_str, pref_str, ccode_str, pcode_str] ; 
  in ifds end
  handle _ => raise Fail "procprf" ; 

(* findcand : 'a vector vector -> int list -> 'a
  find candidate number (within electorate) from codes for 
  party and candidate-within-party *)
fun findcand evec [_, ccode, pcode] = 
  let val cvec = Vector.sub (evec, pcode) ;
  in Vector.sub (cvec, ccode) end ;

(* validprefs' : int -> int list -> int option
  nv = number of valid preferences already seen
  validprefs : int list -> int option
  get sublist of consecutive integers, note [2,3,3,4] -> [2],
  return NONE if full list, SOME l where l is length of sublist *)
fun validprefs' nv (x1 :: x2 :: xs) = 
    if x1+1 = x2 then validprefs' (nv + 1) (x2 :: xs)
    else if x2 = x1 then SOME nv
    else if x2 < x1 then raise Fail "validprefs', x2 < x1"
    else SOME (nv + 1)
  | validprefs' _ [x] = NONE 
  | validprefs' _ [] = raise Fail "validprefs', []"

fun validprefs (xs as 1 :: _) = validprefs' 0 xs
  | validprefs xs = raise Fail "validprefs, no first preference" ;

exception Procvote of int list list ;

type voteid = {batch : int, pindex : int} ;

(* procvote : 'a vector vector -> int list list -> voteid * 'a list
  given a list of preference records for a vote, 
  obtain vote as single list of preferred candidates in order *)
fun procvote evec (plist as (batch :: pindex :: _) :: _) = 
  let val pcs = map (tl o tl) plist ;
    val voteid = {batch = batch, pindex = pindex} ;
    val prefs = map hd pcs ;
    val pv = validprefs prefs ;
    val pcs' = case pv of NONE => pcs | SOME nv => List.take (pcs, nv) ;
    val cands = map (findcand evec) pcs' ;
  in (voteid, cands) end 
  handle _ => raise Procvote plist ;

fun trappend xs ys = List.revAppend (rev xs, ys) ;

fun trconcat (xs :: xss) = trappend xs (trconcat xss)
  | trconcat [] = [] ;

(* getf : string -> substring list
  get list of lines in file, each as a substring *)
fun getf file = 
  let val fs = TextIO.openIn file ;
  val flns = (Substring.tokens (fn ch => ch = #"\n")
    (Substring.all (TextIO.inputAll fs))) ;
  in flns end ;

(* getvf : 'a vector vector -> string list -> (voteid * 'a list) list
  get votes from list of files using candidate vector evec *)
fun getvf evec files = 
  let
    val _ = print "reading vote files\n" ;
    val eleclns = tl (trconcat (map getf files)) ;
    val _ = print "initial processing vote records\n" ;
    val eifs = Gen.maprev procprf eleclns [] ;
    (* simply sorting the lot uses too much memory *)
    (* group by batch, then sort by batch *)
    val _ = print "grouping, sorting (etc) vote records\n" ;
    val gbs = Gen.groupreq hd eifs ;
    val sgbs = Listsort.sort (Gen.trComp (hd o hd) Int.compare) gbs ;
    (* group by batch again, to merge groups of same batch *)
    val ggbs = Gen.groupeq (hd o hd) sgbs ;
    val ggbs' = map List.concat ggbs ;
    (* within batch, sort by pindex and pref *)
    val comp = Gen.trComp tl (Gen.lexComp Int.compare) ;
    val seifs = map (Listsort.sort comp) ggbs' ;
    val geifs = map (Gen.groupeq (fn (_ :: p :: _ ) => p)) seifs ;
    val geifs' = List.concat geifs ;
    val ocls = map (procvote evec) geifs' ;
  in ocls end ;

end ; (* structure Getvotes *)

