
structure Main = struct 
 
val usage = "usage: progname nc ne nv decd elecfile votefile,\n \
  \ nc = number of candidates\n \
  \ ne = number to be elected\n \
  \ nv = number of votes\n \
  \ decd = deceased candidates (eg, _ if none, or 3,7,10)\n \
  \ elecfile = name of electorate file\n \
  \ votefile = name of file containing votes\n" ;

fun do_main (_, args) = 
  let
    val [snc, sne, snv, sdecd, elecfile, votefile] = 
      let 
      in if length args = 6 then args
	else (print usage ; raise Fail "6 arguments required") end ;

    val [nc, ne] = map (Option.valOf o Int.fromString) [snc, sne] 
      handle _ => (print usage ; raise Fail "argument should be integer") ;
    
    val nv = case Int.fromString snv of NONE => ~1 (* read entire votefile *)
	| SOME nv => nv ;

    val decd = map (Option.valOf o Int.fromString) 
      (String.tokens (not o Char.isDigit) sdecd) ;

    (* votefile in the format 
      CarolineLECOUTEUR,JimCOATES,RolandMANDERSON,JaneERREY\n 
      must have extension .cl, other filenames taken to be
      no prefs, then prefs, all as bytes *)
    val (nfv, votes) = 
      case OS.Path.ext votefile of SOME "cl" => 
	  Cand.readclv elecfile votefile nv
	| _ => RPV.readsvotes votefile nv ;

    fun commLot cnt cands = cnt mod length cands 
      handle exn => raise Fail ("commLot: " ^ exnMessage exn) ;

    val info = Count.prefCount (nc, ne, nfv, decd, commLot, votes) ;

    val () = Gen.printElec info ;

    val base = OS.Path.base votefile ;
    val () = Tab.prscr info base ;
    val () = Tab.prdist info base ;

  in () end ;

end ; (* structure Main *)

(* 
mml
load "Main" ;
val [nc, ne, nv] = [28, 5, 16] ;
val decd = [] ;
val elecfile = "Brindabella" ;
val votefile = "votes.dmp.cl" ;

val argstr = "28 5 2 _ Brindabella votes2.cl" ;
val [snc, sne, snv, sdecd, elecfile, votefile] = 
  String.tokens Char.isSpace argstr ;
*)

