
structure Ran = struct 

(* create random data *)

(* so can change generator or use standard one *)

fun stdgen () = (Random.newgenseed 1.0) ;

val gen = (stdgen ()) ;

fun mkpops nc = (Random.rangelist (1, 21) (nc, (gen))) ;
fun number i [] = [] 
  | number i (x :: xs) = (i, x) :: number (i+1) xs ;

(* choice : int * ('a * int) list -> 'a * (int * ('a * int) list)
  given pops - list of (candidate, popularity) pairs, and 
  psum - sum of popularities,
  randomly choose one candidate, and return him and new (psum, pops) *)
fun choice' (psum, pops) ran = 
  let 
    fun find ran' ((cp as (cand, pop)) :: pops) = 
        if ran' < pop then (cp, pops)
        else let val (cp', pops') = find (ran' - pop) pops
          in (cp', cp :: pops') end ;
    val ((cp as (cand, pop)), pops') = find ran pops ;
  in (cand, (psum - pop, pops')) end ;

fun choice (psum, pops) = 
  let val ran = Random.range (0, psum) (gen) ;
  in choice' (psum, pops) ran end ;

(* choices : int -> int * ('a * int) list -> 'a list 
  gets n choices, where psp is (total weight, list of choices with weights) *)
fun choices 0 psp = []
  | choices n psp = 
    let val (c, psp') = choice psp ;
    in c :: choices (n-1) psp' end ;

fun ranpvote nc psp = 
  let val nprefs = Random.range (1, nc + 1) (gen) ;
  in choices nprefs psp end ;

(* collvotes : (unit -> 'a) -> ('a -> 'b -> 'b) -> 'b -> int -> 'b
  mk - makes a vote, coll - collects a new vote, csf - votes collected so far,
  collects votes using arbitrary creation and collection functions *)
fun collvotes _ _ csf 0 = csf
  | collvotes mk coll csf n = collvotes mk coll (coll (mk ()) csf) (n-1) ;

(* ranpvotes : int -> int -> string list
  returns list of nv votes for nc candidates, of random lengths, and
  where candidates have random popularities *)
fun ranpvotes' nc nv (coll, init) = 
  let
    val pops = number 1 (mkpops nc) ;
    val psum = List.foldr (op +) 0 (map #2 pops) ;
    fun mk () = implode (map chr (ranpvote nc (psum, pops))) ;
  in collvotes mk coll init nv end ;

fun ranpvotes nc nv = 
  let fun coll x xs = x :: xs ;
  in ranpvotes' nc nv (coll, []) end ; 

(* test
val pops = number 1 (mkpops 9) ;
val psum = List.foldr (op +) 0 (map #2 pops) ;
load "MyLib" ;
val rans = MyLib.upto (0, psum - 1) ;
val r = map (#1 o choice' (psum, pops)) rans ;
*)

(* ranperm : generator -> int -> int array -> int -> int list
  random distinct permutation of m numbers taken from 1,...,n,
  using array a (size >= n) as scratch space *)
fun ranperm gen n a m = 
  let open Array ;
    fun f (i, _) = i + 1 ;
    val () = modifyi f (a, 0, SOME n) ;
    (* select one from {a(0), ..., a(k-1)},
      and replace one selected by a(k-1), so next selection will be
      from {a(0), ..., a(k-2)} *)
    fun select k = 
      let val ind = Random.range (0, k) gen ;
        val sel = sub (a, ind) ;
        val () = update (a, ind, sub (a, k-1)) ;
      in sel end ;
    (* select m items from {a(0), ..., a(n-1)}, prepend them to rest *)
    fun selectall n 0 rest = rest
      | selectall n m rest = selectall (n-1) (m-1) (select n :: rest) ;
  in selectall n m [] end ;

(* ranvote : generator -> int -> int array -> int list
  as ranperm, but number m also random from 1,...,n *)
fun ranvote gen n a = 
  let val nprefs = Random.range (1, n) gen ;
  in ranperm gen n a nprefs end ;
  
(* ranvotes : generator -> int -> int -> int list list
  produce nv random votes for nc candidates, tail-recursive *)
fun ranvotes gen nc nv = 
  let val a = Array.array (nc, 0) ;
    fun rvs 0 rest = rest
      | rvs n rest = rvs (n-1) (ranvote gen nc a :: rest) ;
  in rvs nv [] end ;

(* ranvotes : generator -> int -> int -> int list list
  produce nv random votes for nc candidates, non-tail-recursive *)
fun ranvotes' gen nc nv = 
  let val a = Array.array (nc, 0) ;
    fun rvs 0 = []
      | rvs n = (ranvote gen nc a :: rvs (n-1)) ;
  in rvs nv end ;

end ; (* struct Ran *)

